# Copyright 2025 wittenator
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file contains code derived from:
# - jax-fid by Matthias Wright (https://github.com/matthias-wright/jax-fid)
#   Copyright 2021 Matthias Wright, Licensed under Apache License 2.0
#
# The original Inception V3 implementation is from:
# https://github.com/matthias-wright/jax-fid/blob/main/jax_fid/inception.py

import functools
import os
import pickle
import tempfile
from collections.abc import Callable, Iterable
from typing import Any

import flax.linen as nn
import jax
import jax.numpy as jnp
import requests
from flax import nnx
from flax.linen.module import merge_param
from flax.nnx import bridge
from jax import lax
from jax.nn import initializers
from jaxtyping import Array, Float, Int, PRNGKeyArray
from tqdm import tqdm

#######################################
######## FID Utilities.
#######################################


def get_fid_network(dtype: str = "float32", ckpt_dir: str | None = "data") -> nnx.Module:
    # Returns a function with input of images in range [-1, 1], and output of 2048-length activations.
    model = InceptionV3(pretrained=True, dtype=dtype, ckpt_dir=ckpt_dir)
    model = bridge.ToNNX(model, rngs=nnx.Rngs(0))
    bridge.lazy_init(model, jnp.ones((1, 299, 299, 3)))
    return model


#######################################
######## Inception V3 Model.
######## https://github.com/matthias-wright/jax-fid/blob/main/jax_fid/inception.py
#######################################


def download(url: str, ckpt_dir: str | None = "data") -> str:
    name = url[url.rfind("/") + 1 : url.rfind("?")]
    if ckpt_dir is None:
        ckpt_dir = tempfile.gettempdir()
    ckpt_file = os.path.join(ckpt_dir, name)
    if not os.path.exists(ckpt_file):
        print(f'Downloading: "{url[: url.rfind("?")]}" to {ckpt_file}')
        if not os.path.exists(ckpt_dir):
            os.makedirs(ckpt_dir)

        response = requests.get(url, stream=True)
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

        # first create temp file, in case the download fails
        ckpt_file_temp = os.path.join(ckpt_dir, name + ".temp")
        cur_data_len = 0
        with open(ckpt_file_temp, "wb") as file:
            for data in response.iter_content(chunk_size=1024):
                progress_bar.update(len(data))
                file.write(data)
                cur_data_len += len(data)
        progress_bar.close()

        if total_size_in_bytes != 0 and cur_data_len != total_size_in_bytes:
            print("An error occured while downloading, please try again.")
            if os.path.exists(ckpt_file_temp):
                os.remove(ckpt_file_temp)
        else:
            # if download was successful, rename the temp file
            os.rename(ckpt_file_temp, ckpt_file)
    return ckpt_file


def get(dictionary: dict | None, key: str):
    if dictionary is None or key not in dictionary:
        return None
    return dictionary[key]


class InceptionV3(nn.Module):
    """
    InceptionV3 network.
    Reference: https://arxiv.org/abs/1512.00567
    Ported mostly from: https://github.com/pytorch/vision/blob/master/torchvision/models/inception.py

    Attributes:
        include_head (bool): If True, include classifier head.
        num_classes (int): Number of classes.
        pretrained (bool): If True, use pretrained weights.
        transform_input (bool): If True, preprocesses the input according to the method with which it
                                was trained on ImageNet.
        aux_logits (bool): If True, add an auxiliary branch that can improve training.
        dtype (str): Data type.
    """

    include_head: bool = False
    num_classes: int = 1000
    pretrained: bool = False
    transform_input: bool = False
    aux_logits: bool = False
    ckpt_path: str = "https://www.dropbox.com/s/xt6zvlvt22dcwck/inception_v3_weights_fid.pickle?dl=1"
    dtype: str = "float32"
    ckpt_dir: str | None = "data"

    def setup(self):
        if self.pretrained:
            ckpt_file = download(self.ckpt_path, ckpt_dir=self.ckpt_dir)
            self.params_dict = pickle.load(open(ckpt_file, "rb"))
            self.num_classes_ = 1000
        else:
            self.params_dict = None
            self.num_classes_ = self.num_classes

    @nn.compact
    def __call__(
        self,
        x: Float[Array, "batch h w c"],
        train: bool = True,
        rng: PRNGKeyArray = None,
    ) -> Array | tuple[Array, Array]:
        """
        Args:
            x (tensor): Input image, shape [B, H, W, C].
            train (bool): If True, training mode.
            rng (jax.random.PRNGKeyArray): Random seed.
        """
        if rng is None:
            rng = jax.random.PRNGKey(0)
        x = self._transform_input(x)
        x = BasicConv2d(
            out_channels=32,
            kernel_size=(3, 3),
            strides=(2, 2),
            params_dict=get(self.params_dict, "Conv2d_1a_3x3"),
            dtype=self.dtype,
        )(x, train)
        x = BasicConv2d(
            out_channels=32, kernel_size=(3, 3), params_dict=get(self.params_dict, "Conv2d_2a_3x3"), dtype=self.dtype
        )(x, train)
        x = BasicConv2d(
            out_channels=64,
            kernel_size=(3, 3),
            padding=((1, 1), (1, 1)),
            params_dict=get(self.params_dict, "Conv2d_2b_3x3"),
            dtype=self.dtype,
        )(x, train)
        x = nn.max_pool(x, window_shape=(3, 3), strides=(2, 2))
        x = BasicConv2d(
            out_channels=80, kernel_size=(1, 1), params_dict=get(self.params_dict, "Conv2d_3b_1x1"), dtype=self.dtype
        )(x, train)
        x = BasicConv2d(
            out_channels=192, kernel_size=(3, 3), params_dict=get(self.params_dict, "Conv2d_4a_3x3"), dtype=self.dtype
        )(x, train)
        x = nn.max_pool(x, window_shape=(3, 3), strides=(2, 2))
        x = InceptionA(pool_features=32, params_dict=get(self.params_dict, "Mixed_5b"), dtype=self.dtype)(x, train)
        x = InceptionA(pool_features=64, params_dict=get(self.params_dict, "Mixed_5c"), dtype=self.dtype)(x, train)
        x = InceptionA(pool_features=64, params_dict=get(self.params_dict, "Mixed_5d"), dtype=self.dtype)(x, train)
        x = InceptionB(params_dict=get(self.params_dict, "Mixed_6a"), dtype=self.dtype)(x, train)
        x = InceptionC(channels_7x7=128, params_dict=get(self.params_dict, "Mixed_6b"), dtype=self.dtype)(x, train)
        x = InceptionC(channels_7x7=160, params_dict=get(self.params_dict, "Mixed_6c"), dtype=self.dtype)(x, train)
        x = InceptionC(channels_7x7=160, params_dict=get(self.params_dict, "Mixed_6d"), dtype=self.dtype)(x, train)
        x = InceptionC(channels_7x7=192, params_dict=get(self.params_dict, "Mixed_6e"), dtype=self.dtype)(x, train)
        aux = None
        if self.aux_logits and train:
            aux = InceptionAux(
                num_classes=self.num_classes_, params_dict=get(self.params_dict, "AuxLogits"), dtype=self.dtype
            )(x, train)
        x = InceptionD(params_dict=get(self.params_dict, "Mixed_7a"), dtype=self.dtype)(x, train)
        x = InceptionE(avg_pool, params_dict=get(self.params_dict, "Mixed_7b"), dtype=self.dtype)(x, train)
        # Following the implementation by @mseitzer, we use max pooling instead
        # of average pooling here.
        # See: https://github.com/mseitzer/pytorch-fid/blob/master/src/pytorch_fid/inception.py#L320
        x = InceptionE(nn.max_pool, params_dict=get(self.params_dict, "Mixed_7c"), dtype=self.dtype)(x, train)
        x = jnp.mean(x, axis=(1, 2), keepdims=True)
        if not self.include_head:
            return x
        x = nn.Dropout(rate=0.5)(x, deterministic=not train, rng=rng)
        x = jnp.reshape(x, newshape=(x.shape[0], -1))
        x = Dense(features=self.num_classes_, params_dict=get(self.params_dict, "fc"), dtype=self.dtype)(x)
        if self.aux_logits:
            return x, aux
        return x

    def _transform_input(self, x: Float[Array, "batch h w c"]) -> Float[Array, "batch h w c"]:
        if self.transform_input:
            x_ch0 = jnp.expand_dims(x[..., 0], axis=-1) * (0.229 / 0.5) + (0.485 - 0.5) / 0.5
            x_ch1 = jnp.expand_dims(x[..., 1], axis=-1) * (0.224 / 0.5) + (0.456 - 0.5) / 0.5
            x_ch2 = jnp.expand_dims(x[..., 2], axis=-1) * (0.225 / 0.5) + (0.406 - 0.5) / 0.5
            x = jnp.concatenate((x_ch0, x_ch1, x_ch2), axis=-1)
        return x


class Dense(nn.Module):
    features: int
    kernel_init: functools.partial = nn.initializers.lecun_normal()
    bias_init: functools.partial = nn.initializers.zeros
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Float[Array, "batch ..."]) -> Array:
        x = nn.Dense(
            features=self.features,
            kernel_init=self.kernel_init
            if self.params_dict is None
            else lambda *_: jnp.array(self.params_dict["kernel"]),
            bias_init=self.bias_init if self.params_dict is None else lambda *_: jnp.array(self.params_dict["bias"]),
        )(x)
        return x


class BasicConv2d(nn.Module):
    out_channels: int
    kernel_size: int | Iterable[int] = (3, 3)
    strides: Iterable[int] | None = (1, 1)
    padding: str | Iterable[tuple[int, int]] = "valid"
    use_bias: bool = False
    kernel_init: functools.partial = nn.initializers.lecun_normal()
    bias_init: functools.partial = nn.initializers.zeros
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        x = nn.Conv(
            features=self.out_channels,
            kernel_size=self.kernel_size,
            strides=self.strides,
            padding=self.padding,
            use_bias=self.use_bias,
            kernel_init=self.kernel_init
            if self.params_dict is None
            else lambda *_: jnp.array(self.params_dict["conv"]["kernel"]),
            bias_init=self.bias_init
            if self.params_dict is None
            else lambda *_: jnp.array(self.params_dict["conv"]["bias"]),
            dtype=self.dtype,
        )(x)
        if self.params_dict is None:
            x = BatchNorm(epsilon=0.001, momentum=0.1, use_running_average=not train, dtype=self.dtype)(x)
        else:
            x = BatchNorm(
                epsilon=0.001,
                momentum=0.1,
                bias_init=lambda *_: jnp.array(self.params_dict["bn"]["bias"]),
                scale_init=lambda *_: jnp.array(self.params_dict["bn"]["scale"]),
                mean_init=lambda *_: jnp.array(self.params_dict["bn"]["mean"]),
                var_init=lambda *_: jnp.array(self.params_dict["bn"]["var"]),
                use_running_average=not train,
                dtype=self.dtype,
            )(x)
        x = jax.nn.relu(x)
        return x


class InceptionA(nn.Module):
    pool_features: int
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        branch1x1 = BasicConv2d(
            out_channels=64, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch1x1"), dtype=self.dtype
        )(x, train)
        branch5x5 = BasicConv2d(
            out_channels=48, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch5x5_1"), dtype=self.dtype
        )(x, train)
        branch5x5 = BasicConv2d(
            out_channels=64,
            kernel_size=(5, 5),
            padding=((2, 2), (2, 2)),
            params_dict=get(self.params_dict, "branch5x5_2"),
            dtype=self.dtype,
        )(branch5x5, train)

        branch3x3dbl = BasicConv2d(
            out_channels=64, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch3x3dbl_1"), dtype=self.dtype
        )(x, train)
        branch3x3dbl = BasicConv2d(
            out_channels=96,
            kernel_size=(3, 3),
            padding=((1, 1), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3dbl_2"),
            dtype=self.dtype,
        )(branch3x3dbl, train)
        branch3x3dbl = BasicConv2d(
            out_channels=96,
            kernel_size=(3, 3),
            padding=((1, 1), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3dbl_3"),
            dtype=self.dtype,
        )(branch3x3dbl, train)

        branch_pool = avg_pool(x, window_shape=(3, 3), strides=(1, 1), padding=((1, 1), (1, 1)))
        branch_pool = BasicConv2d(
            out_channels=self.pool_features,
            kernel_size=(1, 1),
            params_dict=get(self.params_dict, "branch_pool"),
            dtype=self.dtype,
        )(branch_pool, train)

        output = jnp.concatenate((branch1x1, branch5x5, branch3x3dbl, branch_pool), axis=-1)
        return output


class InceptionB(nn.Module):
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        branch3x3 = BasicConv2d(
            out_channels=384,
            kernel_size=(3, 3),
            strides=(2, 2),
            params_dict=get(self.params_dict, "branch3x3"),
            dtype=self.dtype,
        )(x, train)

        branch3x3dbl = BasicConv2d(
            out_channels=64, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch3x3dbl_1"), dtype=self.dtype
        )(x, train)
        branch3x3dbl = BasicConv2d(
            out_channels=96,
            kernel_size=(3, 3),
            padding=((1, 1), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3dbl_2"),
            dtype=self.dtype,
        )(branch3x3dbl, train)
        branch3x3dbl = BasicConv2d(
            out_channels=96,
            kernel_size=(3, 3),
            strides=(2, 2),
            params_dict=get(self.params_dict, "branch3x3dbl_3"),
            dtype=self.dtype,
        )(branch3x3dbl, train)

        branch_pool = nn.max_pool(x, window_shape=(3, 3), strides=(2, 2))

        output = jnp.concatenate((branch3x3, branch3x3dbl, branch_pool), axis=-1)
        return output


class InceptionC(nn.Module):
    channels_7x7: int
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        branch1x1 = BasicConv2d(
            out_channels=192, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch1x1"), dtype=self.dtype
        )(x, train)

        branch7x7 = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(1, 1),
            params_dict=get(self.params_dict, "branch7x7_1"),
            dtype=self.dtype,
        )(x, train)
        branch7x7 = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(1, 7),
            padding=((0, 0), (3, 3)),
            params_dict=get(self.params_dict, "branch7x7_2"),
            dtype=self.dtype,
        )(branch7x7, train)
        branch7x7 = BasicConv2d(
            out_channels=192,
            kernel_size=(7, 1),
            padding=((3, 3), (0, 0)),
            params_dict=get(self.params_dict, "branch7x7_3"),
            dtype=self.dtype,
        )(branch7x7, train)

        branch7x7dbl = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(1, 1),
            params_dict=get(self.params_dict, "branch7x7dbl_1"),
            dtype=self.dtype,
        )(x, train)
        branch7x7dbl = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(7, 1),
            padding=((3, 3), (0, 0)),
            params_dict=get(self.params_dict, "branch7x7dbl_2"),
            dtype=self.dtype,
        )(branch7x7dbl, train)
        branch7x7dbl = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(1, 7),
            padding=((0, 0), (3, 3)),
            params_dict=get(self.params_dict, "branch7x7dbl_3"),
            dtype=self.dtype,
        )(branch7x7dbl, train)
        branch7x7dbl = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(7, 1),
            padding=((3, 3), (0, 0)),
            params_dict=get(self.params_dict, "branch7x7dbl_4"),
            dtype=self.dtype,
        )(branch7x7dbl, train)
        branch7x7dbl = BasicConv2d(
            out_channels=self.channels_7x7,
            kernel_size=(1, 7),
            padding=((0, 0), (3, 3)),
            params_dict=get(self.params_dict, "branch7x7dbl_5"),
            dtype=self.dtype,
        )(branch7x7dbl, train)

        branch_pool = avg_pool(x, window_shape=(3, 3), strides=(1, 1), padding=((1, 1), (1, 1)))
        branch_pool = BasicConv2d(
            out_channels=192, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch_pool"), dtype=self.dtype
        )(branch_pool, train)

        output = jnp.concatenate((branch1x1, branch7x7, branch7x7dbl, branch_pool), axis=-1)
        return output


class InceptionD(nn.Module):
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        branch3x3 = BasicConv2d(
            out_channels=192, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch3x3_1"), dtype=self.dtype
        )(x, train)
        branch3x3 = BasicConv2d(
            out_channels=320,
            kernel_size=(3, 3),
            strides=(2, 2),
            params_dict=get(self.params_dict, "branch3x3_2"),
            dtype=self.dtype,
        )(branch3x3, train)

        branch7x7x3 = BasicConv2d(
            out_channels=192, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch7x7x3_1"), dtype=self.dtype
        )(x, train)
        branch7x7x3 = BasicConv2d(
            out_channels=192,
            kernel_size=(1, 7),
            padding=((0, 0), (3, 3)),
            params_dict=get(self.params_dict, "branch7x7x3_2"),
            dtype=self.dtype,
        )(branch7x7x3, train)
        branch7x7x3 = BasicConv2d(
            out_channels=192,
            kernel_size=(7, 1),
            padding=((3, 3), (0, 0)),
            params_dict=get(self.params_dict, "branch7x7x3_3"),
            dtype=self.dtype,
        )(branch7x7x3, train)
        branch7x7x3 = BasicConv2d(
            out_channels=192,
            kernel_size=(3, 3),
            strides=(2, 2),
            params_dict=get(self.params_dict, "branch7x7x3_4"),
            dtype=self.dtype,
        )(branch7x7x3, train)

        branch_pool = nn.max_pool(x, window_shape=(3, 3), strides=(2, 2))

        output = jnp.concatenate((branch3x3, branch7x7x3, branch_pool), axis=-1)
        return output


class InceptionE(nn.Module):
    pooling: Callable
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        branch1x1 = BasicConv2d(
            out_channels=320, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch1x1"), dtype=self.dtype
        )(x, train)

        branch3x3 = BasicConv2d(
            out_channels=384, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch3x3_1"), dtype=self.dtype
        )(x, train)
        branch3x3_a = BasicConv2d(
            out_channels=384,
            kernel_size=(1, 3),
            padding=((0, 0), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3_2a"),
            dtype=self.dtype,
        )(branch3x3, train)
        branch3x3_b = BasicConv2d(
            out_channels=384,
            kernel_size=(3, 1),
            padding=((1, 1), (0, 0)),
            params_dict=get(self.params_dict, "branch3x3_2b"),
            dtype=self.dtype,
        )(branch3x3, train)
        branch3x3 = jnp.concatenate((branch3x3_a, branch3x3_b), axis=-1)

        branch3x3dbl = BasicConv2d(
            out_channels=448, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch3x3dbl_1"), dtype=self.dtype
        )(x, train)
        branch3x3dbl = BasicConv2d(
            out_channels=384,
            kernel_size=(3, 3),
            padding=((1, 1), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3dbl_2"),
            dtype=self.dtype,
        )(branch3x3dbl, train)
        branch3x3dbl_a = BasicConv2d(
            out_channels=384,
            kernel_size=(1, 3),
            padding=((0, 0), (1, 1)),
            params_dict=get(self.params_dict, "branch3x3dbl_3a"),
            dtype=self.dtype,
        )(branch3x3dbl, train)
        branch3x3dbl_b = BasicConv2d(
            out_channels=384,
            kernel_size=(3, 1),
            padding=((1, 1), (0, 0)),
            params_dict=get(self.params_dict, "branch3x3dbl_3b"),
            dtype=self.dtype,
        )(branch3x3dbl, train)
        branch3x3dbl = jnp.concatenate((branch3x3dbl_a, branch3x3dbl_b), axis=-1)

        branch_pool = self.pooling(x, window_shape=(3, 3), strides=(1, 1), padding=((1, 1), (1, 1)))
        branch_pool = BasicConv2d(
            out_channels=192, kernel_size=(1, 1), params_dict=get(self.params_dict, "branch_pool"), dtype=self.dtype
        )(branch_pool, train)

        output = jnp.concatenate((branch1x1, branch3x3, branch3x3dbl, branch_pool), axis=-1)
        return output


class InceptionAux(nn.Module):
    num_classes: int
    kernel_init: functools.partial = nn.initializers.lecun_normal()
    bias_init: functools.partial = nn.initializers.zeros
    params_dict: dict = None
    dtype: str = "float32"

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> Array:
        x = avg_pool(x, window_shape=(5, 5), strides=(3, 3))
        x = BasicConv2d(
            out_channels=128, kernel_size=(1, 1), params_dict=get(self.params_dict, "conv0"), dtype=self.dtype
        )(x, train)
        x = BasicConv2d(
            out_channels=768, kernel_size=(5, 5), params_dict=get(self.params_dict, "conv1"), dtype=self.dtype
        )(x, train)
        x = jnp.mean(x, axis=(1, 2))
        x = jnp.reshape(x, newshape=(x.shape[0], -1))
        x = Dense(features=self.num_classes, params_dict=get(self.params_dict, "fc"), dtype=self.dtype)(x)
        return x


def _absolute_dims(rank: int, dims: Iterable[int]) -> tuple[int, ...]:
    return tuple([rank + dim if dim < 0 else dim for dim in dims])


class BatchNorm(nn.Module):
    use_running_average: bool | None = None
    axis: int = -1
    momentum: float = 0.99
    epsilon: float = 1e-5
    dtype: jnp.dtype = jnp.float32
    use_bias: bool = True
    use_scale: bool = True
    bias_init: Callable[[PRNGKeyArray, tuple[int, ...], jnp.dtype], Array] = initializers.zeros
    scale_init: Callable[[PRNGKeyArray, tuple[int, ...], jnp.dtype], Array] = initializers.ones
    mean_init: Callable[[tuple[int, ...]], Array] = lambda s: jnp.zeros(s, jnp.float32)
    var_init: Callable[[tuple[int, ...]], Array] = lambda s: jnp.ones(s, jnp.float32)
    axis_name: str | None = None
    axis_index_groups: Any = None

    @nn.compact
    def __call__(self, x: jnp.ndarray, use_running_average: bool | None = None) -> jnp.ndarray:
        use_running_average = merge_param("use_running_average", self.use_running_average, use_running_average)
        x = jnp.asarray(x, jnp.float32)
        axis = self.axis if isinstance(self.axis, tuple) else (self.axis,)
        axis = _absolute_dims(x.ndim, axis)
        feature_shape = tuple(d if i in axis else 1 for i, d in enumerate(x.shape))
        reduced_feature_shape = tuple(d for i, d in enumerate(x.shape) if i in axis)
        reduction_axis = tuple(i for i in range(x.ndim) if i not in axis)

        # see NOTE above on initialization behavior
        initializing = self.is_mutable_collection("params")

        ra_mean = self.variable("batch_stats", "mean", self.mean_init, reduced_feature_shape)
        ra_var = self.variable("batch_stats", "var", self.var_init, reduced_feature_shape)

        if use_running_average:
            mean, var = ra_mean.value, ra_var.value
        else:
            mean = jnp.mean(x, axis=reduction_axis, keepdims=False)
            mean2 = jnp.mean(lax.square(x), axis=reduction_axis, keepdims=False)
            if self.axis_name is not None and not initializing:
                concatenated_mean = jnp.concatenate([mean, mean2])
                mean, mean2 = jnp.split(
                    lax.pmean(concatenated_mean, axis_name=self.axis_name, axis_index_groups=self.axis_index_groups), 2
                )
            var = mean2 - lax.square(mean)

            if not initializing:
                ra_mean.value = self.momentum * ra_mean.value + (1 - self.momentum) * mean
                ra_var.value = self.momentum * ra_var.value + (1 - self.momentum) * var

        y = x - mean.reshape(feature_shape)
        mul = lax.rsqrt(var + self.epsilon)
        if self.use_scale:
            scale = self.param("scale", self.scale_init, reduced_feature_shape).reshape(feature_shape)
            mul = mul * scale
        y = y * mul
        if self.use_bias:
            bias = self.param("bias", self.bias_init, reduced_feature_shape).reshape(feature_shape)
            y = y + bias
        return jnp.asarray(y, self.dtype)


def pool(
    inputs: jnp.ndarray,
    init: float,
    reduce_fn: Callable,
    window_shape: tuple[int, ...],
    strides: tuple[int, ...] | None,
    padding: str | Iterable[tuple[int, int]],
) -> jnp.ndarray:
    strides = strides or (1,) * len(window_shape)
    assert len(window_shape) == len(strides), f"len({window_shape}) == len({strides})"
    strides = (1,) + strides + (1,)
    dims = (1,) + window_shape + (1,)

    is_single_input = False
    if inputs.ndim == len(dims) - 1:
        # add singleton batch dimension because lax.reduce_window always
        # needs a batch dimension.
        inputs = inputs[None]
        is_single_input = True

    assert inputs.ndim == len(dims), f"len({inputs.shape}) != len({dims})"
    if not isinstance(padding, str):
        padding = tuple(map(tuple, padding))
        assert len(padding) == len(window_shape), (
            f"padding {padding} must specify pads for same number of dims as window_shape {window_shape}"
        )
        assert all([len(x) == 2 for x in padding]), f"each entry in padding {padding} must be length 2"
        padding = ((0, 0),) + padding + ((0, 0),)
    y = jax.lax.reduce_window(inputs, init, reduce_fn, dims, strides, padding)
    if is_single_input:
        y = jnp.squeeze(y, axis=0)
    return y


def avg_pool(
    inputs: jnp.ndarray,
    window_shape: tuple[int, int],
    strides: tuple[int, int] | None = None,
    padding: str | Iterable[tuple[int, int]] = "VALID",
) -> jnp.ndarray:
    assert inputs.ndim == 4
    assert len(window_shape) == 2

    y = pool(inputs, 0.0, jax.lax.add, window_shape, strides, padding)
    ones = jnp.ones(shape=(1, inputs.shape[1], inputs.shape[2], 1)).astype(inputs.dtype)
    counts = jax.lax.conv_general_dilated(
        ones,
        jnp.expand_dims(jnp.ones(window_shape).astype(inputs.dtype), axis=(-2, -1)),
        window_strides=(1, 1),
        padding=((1, 1), (1, 1)),
        dimension_numbers=nn.linear._conv_dimension_numbers(ones.shape),
        feature_group_count=1,
    )
    y = y / counts
    return y



"""Test for Fréchet Inception Distance (FID) metric implementation."""

import logging
import os
import pickle
import tempfile
import time
from pathlib import Path

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = "0.3"  # Disable preallocation to use only 50% of GPU memory
import jax
import jax.numpy as jnp
import numpy as np
import torch
import torchmetrics
import torchvision
import torchvision.transforms as T
import torchvision.transforms.v2 as T2
from jaxtyping import install_import_hook

with install_import_hook("scripts", "typeguard.typechecked"):
    from fidax.fid import FrechetInceptionDistance

# activate fp64
jax.config.update("jax_enable_x64", True)

logger = logging.getLogger("fidax.tests.timing")


def test_fid_equivalence_to_torchmetrics() -> None:
    """Test JAX FID implementation against torchmetrics for equivalence, using batched updates for larger N."""
    # Generate random fake and real images in [-1, 1], shape [N, 299, 299, 3] for jax
    N = 128  # Larger N
    batch_size = 32
    np.random.seed(0)  # For reproducibility
    fake_imgs = np.random.uniform(size=(N, 299, 299, 3), low=-1, high=1)
    fake_imgs = fake_imgs / 2 + 0.5
    real_imgs = np.random.uniform(size=(N, 299, 299, 3), low=-1, high=1)
    real_imgs = real_imgs / 2 + 0.5

    # Torchmetrics expects [N, C, H, W] in [0, 1]
    fake_imgs_torch = torch.tensor(np.array(fake_imgs)).permute(0, 3, 1, 2)
    real_imgs_torch = torch.tensor(np.array(real_imgs)).permute(0, 3, 1, 2)

    # PyTorch FID (use CUDA if available)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    fake_imgs_torch = fake_imgs_torch.to(device)
    real_imgs_torch = real_imgs_torch.to(device)
    fid_torch = torchmetrics.image.FrechetInceptionDistance(feature=2048, normalize=True).to(device)
    fid_torch = fid_torch.set_dtype(torch.float64)
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_torch.update(real_imgs_torch[i : i + batch_size], real=True)
        fid_torch.update(fake_imgs_torch[i : i + batch_size], real=False)
    fid_torch_score = fid_torch.compute().item()
    t_torch = time.perf_counter() - t0
    # clear CUDA memory
    torch.cuda.empty_cache()

    # JAX FrechetInceptionDistance (NNX style)
    fid_jax = FrechetInceptionDistance()
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax.update(real_imgs[i : i + batch_size], True)
        fid_jax.update(fake_imgs[i : i + batch_size], False)
    jax_score = float(fid_jax.compute())
    t_jax = time.perf_counter() - t0

    logger.info(
        "timing equivalence: torchmetrics=%.3fs | jax=%.3fs | speedup x%.2f",
        t_torch,
        t_jax,
        (t_torch / max(t_jax, 1e-9)),
    )

    # Allow a small tolerance due to possible implementation/model differences
    assert np.allclose(jax_score, fid_torch_score, rtol=1e-1, atol=1e-1), (
        f"JAX FID {jax_score} vs Torchmetrics {fid_torch_score}"
    )

    # Test reset functionality
    fid_jax.reset()
    assert fid_jax.real_count == 0
    assert fid_jax.fake_count == 0


def test_fid_with_precomputed_stats() -> None:
    """Test JAX FID implementation with pre-computed real statistics, using batched updates."""
    # Generate random fake images and pre-compute real stats
    N = 1024
    batch_size = 64
    np.random.seed(0)  # For reproducibility
    fake_imgs = np.random.uniform(size=(N, 299, 299, 3), low=-1, high=1)
    fake_imgs = fake_imgs / 2 + 0.5
    real_imgs = np.random.uniform(size=(N, 299, 299, 3), low=-1, high=1)
    real_imgs = real_imgs / 2 + 0.5

    # First calculate regular FID to get real stats (batched)
    fid_jax = FrechetInceptionDistance()
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax.update(real_imgs[i : i + batch_size], True)
    warmup = time.perf_counter() - t0

    # Extract pre-computed stats using the metric helper
    mu_real, sigma_real = fid_jax.get_real_stats()
    real_stats = {"mu": mu_real, "sigma": sigma_real}

    # Now create new FID with pre-computed stats (batched)
    fid_jax_precomputed = FrechetInceptionDistance(real_stats=real_stats)
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax_precomputed.update(fake_imgs[i : i + batch_size], False)
    jax_score_precomputed = float(fid_jax_precomputed.compute())
    t_precomp = time.perf_counter() - t0

    # Update original FID with fake images and compute (batched)
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax.update(fake_imgs[i : i + batch_size], False)
    jax_score = float(fid_jax.compute())
    t_full = time.perf_counter() - t0

    logger.info(
        "timing precomp: warmup=%.3fs | precomp=%.3fs | full=%.3fs",
        warmup,
        t_precomp,
        t_full,
    )

    # Both scores should be identical
    assert np.allclose(jax_score, jax_score_precomputed, rtol=1e-5, atol=1e-5), (
        f"Standard FID {jax_score} vs Pre-computed stats FID {jax_score_precomputed}"
    )


def test_fid_on_cifar10_real_vs_modified() -> None:
    """Test FID on CIFAR-10 real images vs. a modified version (e.g., noisy), using batched updates."""
    # Download CIFAR-10 and select a small subset for speed
    transform = T.Compose(
        [
            T.ToTensor(),
        ]
    )
    cifar10 = torchvision.datasets.CIFAR10(root="/tmp/cifar10", train=False, download=True, transform=transform)
    N = 9984
    batch_size = 64
    real_imgs_torch = torch.stack([cifar10[i][0] for i in range(N)])  # [N, C, H, W], [0,1]

    # Create a modified version: add Gaussian noise
    noise = torch.randn_like(real_imgs_torch) * 0.1
    fake_imgs_torch = (real_imgs_torch + noise).clamp(0, 1)

    # save one image for visual inspection
    torchvision.utils.save_image(fake_imgs_torch[0], "./fake_cifar10_image.png")
    # print max and min values
    logger.info(f"Fake image max: {fake_imgs_torch[0].max().item()}")

    # PyTorch FID (use CUDA if available)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    fake_imgs_torch = fake_imgs_torch.to(device)
    real_imgs_torch = real_imgs_torch.to(device)
    fid_torch = torchmetrics.image.FrechetInceptionDistance(feature=2048, normalize=True).to(device)
    fid_torch = fid_torch.set_dtype(torch.float64)
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_torch.update(real_imgs_torch[i : i + batch_size], real=True)
        fid_torch.update(fake_imgs_torch[i : i + batch_size], real=False)
    fid_torch_score = fid_torch.compute().item()
    t_torch = time.perf_counter() - t0
    # Move tensors back to CPU for further processing if needed
    fake_imgs_torch = fake_imgs_torch.cpu()
    real_imgs_torch = real_imgs_torch.cpu()
    # clear CUDA memory
    torch.cuda.empty_cache()

    # Convert to JAX format: [N, 299, 299, 3]
    real_imgs = real_imgs_torch.permute(0, 2, 3, 1).numpy()
    fake_imgs = fake_imgs_torch.permute(0, 2, 3, 1).numpy()
    real_imgs = jnp.array(real_imgs)
    fake_imgs = jnp.array(fake_imgs)

    # JAX FID (batched)
    fid_jax = FrechetInceptionDistance()
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax.update(real_imgs[i : i + batch_size], True)
        fid_jax.update(fake_imgs[i : i + batch_size], False)
    jax_score = float(fid_jax.compute())
    t_jax = time.perf_counter() - t0

    logger.info(
        "timing CIFAR noisy: torchmetrics=%.3fs | jax=%.3fs | speedup x%.2f",
        t_torch,
        t_jax,
        (t_torch / max(t_jax, 1e-9)),
    )

    # log the scores
    logger.info("FID Scores: JAX=%.3f | Torchmetrics=%.3f", jax_score, fid_torch_score)

    # FID should be > 0 (since fake is noisy version of real)
    assert jax_score > 0.0
    assert fid_torch_score > 0.0
    # The two implementations should be close
    assert np.allclose(jax_score, fid_torch_score, rtol=1e-1, atol=1e-1), (
        f"JAX FID {jax_score} vs Torchmetrics {fid_torch_score}"
    )


def test_fid_on_cifar10_real_vs_random_erasing() -> None:
    """Test FID on CIFAR-10 real images vs. a RandomErasing-augmented version, using batched updates."""
    # Download CIFAR-10 and select a small subset for speed
    transform = T.Compose(
        [
            T.ToTensor(),
        ]
    )
    cifar10 = torchvision.datasets.CIFAR10(root="/tmp/cifar10", train=False, download=True, transform=transform)
    N = 9984
    batch_size = 64
    real_imgs_torch = torch.stack([cifar10[i][0] for i in range(N)])  # [N, C, H, W], [0,1]

    # Create a modified version: apply RandomErasing
    random_erasing = T2.RandomErasing(p=1.0, scale=(0.2, 0.4), ratio=(0.3, 3.3))
    fake_imgs_torch = torch.stack([random_erasing(img) for img in real_imgs_torch])
    # save one image for visual inspection
    torchvision.utils.save_image(fake_imgs_torch[0], "./erased_fake_cifar10_image.png")
    # print max and min values
    logger.info(f"Fake image max: {fake_imgs_torch[0].max().item()}")

    # PyTorch FID (use CUDA if available)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    fake_imgs_torch = fake_imgs_torch.to(device)
    real_imgs_torch = real_imgs_torch.to(device)
    fid_torch = torchmetrics.image.FrechetInceptionDistance(feature=2048, normalize=True).to(device)
    fid_torch = fid_torch.set_dtype(torch.float64)
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_torch.update(real_imgs_torch[i : i + batch_size], real=True)
        fid_torch.update(fake_imgs_torch[i : i + batch_size], real=False)
    fid_torch_score = fid_torch.compute().item()
    t_torch = time.perf_counter() - t0
    # Move tensors back to CPU for further processing if needed
    fake_imgs_torch = fake_imgs_torch.cpu()
    real_imgs_torch = real_imgs_torch.cpu()
    # clear CUDA memory
    torch.cuda.empty_cache()

    # Convert to JAX format: [N, 299, 299, 3], [-1, 1]
    real_imgs = real_imgs_torch.permute(0, 2, 3, 1).numpy()
    fake_imgs = fake_imgs_torch.permute(0, 2, 3, 1).numpy()
    real_imgs = jnp.array(real_imgs)
    fake_imgs = jnp.array(fake_imgs)

    # JAX FID (batched)
    fid_jax = FrechetInceptionDistance()
    t0 = time.perf_counter()
    for i in range(0, N, batch_size):
        fid_jax.update(real_imgs[i : i + batch_size], True)
        fid_jax.update(fake_imgs[i : i + batch_size], False)
    jax_score = float(fid_jax.compute())
    t_jax = time.perf_counter() - t0

    logger.info(
        "timing CIFAR erase: torchmetrics=%.3fs | jax=%.3fs | speedup x%.2f",
        t_torch,
        t_jax,
        (t_torch / max(t_jax, 1e-9)),
    )

    # log the scores
    logger.info("FID Scores: JAX=%.3f | Torchmetrics=%.3f", jax_score, fid_torch_score)

    # FID should be > 0 (since fake is RandomErasing version of real)
    assert jax_score > 0.0
    assert fid_torch_score > 0.0
    # The two implementations should be close
    assert np.allclose(jax_score, fid_torch_score, rtol=1e-1, atol=1e-1), (
        f"JAX FID {jax_score} vs Torchmetrics {fid_torch_score}"
    )


def test_fid_weights_cache_dir_uses_custom_path(tmp_path, monkeypatch) -> None:
    """Verify that passing weights_cache_dir stores weights in that directory and passes it to downloader.

    We monkeypatch fidax.inception.download to avoid network and to assert the ckpt_dir used.
    """
    seen = {}

    def fake_download(_url: str, ckpt_dir: str | None = "data") -> str:
        # Record the directory the metric passed to the downloader
        seen["ckpt_dir"] = ckpt_dir
        # Emulate download logic: if no dir provided, use system temp dir
        target_dir = Path(ckpt_dir) if ckpt_dir is not None else Path(tempfile.gettempdir())
        target_dir.mkdir(parents=True, exist_ok=True)
        # Create an empty, but valid pickle file for weights
        fname = "inception_v3_weights_fid.pickle"
        fpath = target_dir / fname
        if not fpath.exists():
            with fpath.open("wb") as f:
                pickle.dump({}, f)
        return str(fpath)

    # Patch the download function used by the Inception model
    monkeypatch.setattr("fidax.inception.download", fake_download)

    cache_dir = tmp_path / "weights-cache"

    # Create the metric with a custom cache directory; this triggers model init and our fake download
    _ = FrechetInceptionDistance(weights_cache_dir=str(cache_dir))

    # The downloader should have been called with our custom directory
    assert seen.get("ckpt_dir") == str(cache_dir)
    # And the weights file should exist in that directory
    assert (cache_dir / "inception_v3_weights_fid.pickle").exists()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict

import jax
import jax.numpy as jnp
from flax import nnx
from flax.nnx.training.metrics import MetricState

from fidax.inception import get_fid_network

if TYPE_CHECKING:
    from jaxtyping import Array, ArrayLike, Float, Int


class Stats(TypedDict):
    """Typed container for real/fake distribution statistics used by FID."""

    mu: Float[Array, "feat"]
    sigma: Float[Array, "feat feat"]


def _extract_activations(model, imgs: Float[Array, "batch h w c"]) -> Float[Array, "batch 2048"]:
    """Run Inception and return [N, 2048] activations (model's dtype)."""
    # Handle grayscale images by repeating channels to RGB
    if imgs.shape[-1] == 1:
        imgs = jnp.tile(imgs, (1, 1, 1, 3))

    imgs = jax.image.resize(imgs, (imgs.shape[0], 299, 299, 3), method="bilinear")
    acts = model(imgs, train=False)[..., 0, 0, :]
    return acts


class FrechetInceptionDistance(nnx.Metric):
    """Fréchet Inception Distance (FID) metric implemented in JAX."""

    def __init__(
        self,
        metric_dtype: jnp.dtype = jnp.float64,
        model_dtype: str = "float32",
        real_stats: Stats | None = None,
        weights_cache_dir: str | None = "data",
    ) -> None:
        """Initialize FID metric.

        Args:
            metric_dtype: Data type for computations for model execution and metric computation (default: jnp.float64)
            model_dtype: Data type for the model (default: jnp.float32)
            real_stats: Pre-computed statistics for real images (mu and sigma)
            weights_cache_dir: Directory to cache/download Inception weights
        """
        self.real_stats = real_stats
        self.metric_dtype = metric_dtype
        self.model_dtype = model_dtype
        self.weights_cache_dir = weights_cache_dir

        # Initialize feature extractor
        self.model = get_fid_network(dtype=model_dtype, ckpt_dir=self.weights_cache_dir)

        # Dimension of Inception feature activations
        self._feat_dim = 2048

        # Streaming accumulators on DEVICE (constant memory wrt number of samples)
        # Maintain (count, mean[D], M2[D,D]) per split as JAX arrays
        self._real_n = MetricState(jnp.array(0, dtype=jnp.int32))
        self._real_mean = MetricState(jnp.zeros((self._feat_dim,), dtype=self.metric_dtype))
        self._real_M2 = MetricState(jnp.zeros((self._feat_dim, self._feat_dim), dtype=self.metric_dtype))

        self._fake_n = MetricState(jnp.array(0, dtype=jnp.int32))
        self._fake_mean = MetricState(jnp.zeros((self._feat_dim,), dtype=self.metric_dtype))
        self._fake_M2 = MetricState(jnp.zeros((self._feat_dim, self._feat_dim), dtype=self.metric_dtype))

    @nnx.jit(static_argnames=("real",))
    def update(self, imgs: Float[ArrayLike, "batch h w c"], real: bool, **kwargs: Any) -> None:  # noqa: ARG002, D417, RUF100
        """Update the metric with new images.

        Args:
            imgs: Input images, shape [N, H, W, C], range [0, 1]
            real: Whether the images are real (True) or generated (False)
            kwargs: Additional keyword arguments (unused)

        Note:
            This method computes activations on the accelerator and aggregates
            statistics on the host, avoiding large device-resident buffers.
        """
        imgs = imgs.astype(self.model_dtype) * 2.0 - 1.0  # Scale to [-1, 1]
        # Run model forward on device and update accumulators on device as well
        acts = _extract_activations(self.model, imgs).astype(self.metric_dtype)  # [B, D]
        nb = acts.shape[0]
        mb = jnp.mean(acts, axis=0)
        # Compute batch second moment and convert to centered sum of squares (M2b)
        S2b = acts.T @ acts  # (D,D)
        nb_f = jnp.array(nb, dtype=self.metric_dtype)
        M2b = S2b - nb_f * jnp.outer(mb, mb)  # (D,D)

        # Select current split accumulators
        n = self._real_n.value if real else self._fake_n.value
        mean = self._real_mean.value if real else self._fake_mean.value
        M2 = self._real_M2.value if real else self._fake_M2.value

        # Combine using Chan's formula (works also when n == 0)
        n_f = n.astype(self.metric_dtype)
        nb_f = jnp.array(nb, dtype=self.metric_dtype)
        n_new = n + jnp.array(nb, dtype=n.dtype)
        n_new_f = n_f + nb_f
        mean_new = jnp.where(n_new_f > 0, (mean * n_f + mb * nb_f) / n_new_f, jnp.zeros_like(mean))
        delta = mb - mean
        M2_new = (
            M2
            + M2b
            + jnp.outer(delta, delta)
            * jnp.where(n_new_f > 0, (n_f * nb_f) / n_new_f, jnp.array(0, dtype=self.metric_dtype))
        )

        if real:
            self._real_n.value = n_new
            self._real_mean.value = mean_new
            self._real_M2.value = M2_new
        else:
            self._fake_n.value = n_new
            self._fake_mean.value = mean_new
            self._fake_M2.value = M2_new

    @nnx.jit
    def compute(self) -> float:
        """Compute the FID score between real and fake image distributions."""
        # Derive stats for fake (device)
        mu1, sigma1 = self._stats_from_accumulators_jax(self._fake_n.value, self._fake_mean.value, self._fake_M2.value)

        # Derive stats for real, or use precomputed
        if self.real_stats is not None:
            mu2 = self.real_stats["mu"]
            sigma2 = self.real_stats["sigma"]
        else:
            mu2, sigma2 = self._stats_from_accumulators_jax(
                self._real_n.value, self._real_mean.value, self._real_M2.value
            )

        return self._fid_from_stats(mu1, sigma1, mu2, sigma2)

    @staticmethod
    @jax.jit
    def _fid_from_stats(
        mu1: Float[Array, "feat"],
        sigma1: Float[Array, "feat feat"],
        mu2: Float[Array, "feat"],
        sigma2: Float[Array, "feat feat"],
    ) -> Float[Array, ""]:
        """Compute FID score from distribution statistics.

        Adapted from https://github.com/Lightning-AI/torchmetrics/blob/27e1dbe39ac50d6c84f72a16afbb7bf1eb19221e/src/torchmetrics/image/fid.py
        """
        sigma1 = sigma1 + jnp.eye(sigma1.shape[0]) * 1e-6  # Add small value for numerical stability
        sigma2 = sigma2 + jnp.eye(sigma2.shape[0]) * 1e-6  # Add small value for numerical stability
        diff = jnp.sum((mu1 - mu2) ** 2, axis=-1)
        traces = jnp.trace(sigma1) + jnp.trace(sigma2)

        eigvals = jnp.linalg.eigvals(sigma1 @ sigma2)
        covmean = jnp.sum(jnp.sqrt(jnp.abs(eigvals)).real)
        return diff + traces - 2 * covmean

    def reset(self) -> None:
        """Reset the metric state."""
        self._real_n.value = jnp.array(0, dtype=jnp.int32)
        self._real_mean.value = jnp.zeros((self._feat_dim,), dtype=self.metric_dtype)
        self._real_M2.value = jnp.zeros((self._feat_dim, self._feat_dim), dtype=self.metric_dtype)
        self._fake_n.value = jnp.array(0, dtype=jnp.int32)
        self._fake_mean.value = jnp.zeros((self._feat_dim,), dtype=self.metric_dtype)
        self._fake_M2.value = jnp.zeros((self._feat_dim, self._feat_dim), dtype=self.metric_dtype)

    # --------------------
    # Helper methods (host-side)
    # --------------------
    def _stats_from_accumulators_jax(
        self,
        n: Int[Array, ""],
        mean: Float[Array, "feat"],
        M2: Float[Array, "feat feat"],
    ) -> tuple[Float[Array, "feat"], Float[Array, "feat feat"]]:
        """Return (mu, sigma) from accumulators on device; ddof=1 if n>1, else zeros."""
        n = jnp.asarray(n)
        n_f = n.astype(self.metric_dtype)
        mu = mean
        sigma = jnp.where(n >= 2, M2 / jnp.maximum(n_f - 1.0, 1.0), jnp.zeros_like(M2))
        return mu, sigma

    # Public helpers for tests and external use
    def get_real_stats(self) -> tuple[Float[Array, "feat"], Float[Array, "feat feat"]]:
        return self._stats_from_accumulators_jax(self._real_n.value, self._real_mean.value, self._real_M2.value)

    def get_fake_stats(self) -> tuple[Float[Array, "feat"], Float[Array, "feat feat"]]:
        return self._stats_from_accumulators_jax(self._fake_n.value, self._fake_mean.value, self._fake_M2.value)

    @property
    def real_count(self) -> int:
        return int(self._real_n.value)

    @property
    def fake_count(self) -> int:
        return int(self._fake_n.value)
